# pyOCD debugger
# Copyright (c) 2021 Huada Semiconductor Corporation
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile


class DBGMCU:
    STPCTL = 0xE0042020
    STPCTL_VALUE = 0x3FFB

    TRACECTL = 0xE0042024
    TRACECTL_VALUE = 0x0


FLASH_ALGO = { 'load_address' : 0x20000000,
               'instructions' : [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4770ba40, 0x4770ba40, 0x4770bac0, 0x4770bac0, 0xf000b510, 0x2000f833, 0xb510bd10, 0xf836f000,
    0xbd102000, 0xb876f000, 0xb898f000, 0xb836f000, 0x4603b5f0, 0xe00c2400, 0x0514f1a4, 0xd2022d03,
    0x1d1b1d12, 0xcb20e004, 0x42b56816, 0x1d12d104, 0xebb41c64, 0xd3ef0f91, 0x2300461d, 0x0103f001,
    0xf815e005, 0x5cd76b01, 0xd10242be, 0x42991c5b, 0xeb00d8f7, 0x44180084, 0x0000bdf0, 0xf2404849,
    0x60011123, 0x2110f243, 0x47706001, 0x20004945, 0x6008310c, 0x1e404943, 0x47706008, 0x4f41b5f8,
    0x23012600, 0x9600370c, 0x683b603b, 0x7398f443, 0xf8df603b, 0xf649c0f0, 0xf10c4540, 0x1d3c0c14,
    0x60036813, 0xe0049600, 0x1c5b9b00, 0x42ab9300, 0x6823d219, 0xd5f706db, 0xf8dce005, 0xf0433000,
    0xf8cc0310, 0x68233000, 0xd4f606db, 0x1f091d00, 0x29041d12, 0x603ed2e4, 0xe0069600, 0x1c409800,
    0x42a89000, 0x2001d301, 0x6820bdf8, 0xd5f505c0, 0xbdf82000, 0xb5304a23, 0x21012000, 0x6011320c,
    0xf4416811, 0x601171a8, 0x60004603, 0x1d114c1e, 0x1c40e004, 0xd30142a0, 0xbd302001, 0x05ed680d,
    0x4818d5f7, 0x680c3014, 0xd50406e4, 0xf0446804, 0x60040410, 0x6013e7f7, 0xbd302000, 0xb5304a11,
    0x23012100, 0x6013320c, 0xf4436813, 0x601373a0, 0x6001460b, 0x1d104c0c, 0x1c49e004, 0xd30142a1,
    0xbd302001, 0x05ed6805, 0x4906d5f7, 0x68043114, 0xd50406e4, 0xf044680c, 0x600c0410, 0x6013e7f7,
    0xbd302000, 0x40010400, 0x00061a80, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000031,
    'pc_unInit': 0x2000003b,
    'pc_program_page': 0x2000004d,
    'pc_erase_sector': 0x20000049,
    'pc_eraseAll': 0x20000045,

    'static_base' : 0x20000000 + 0x00000020 + 0x000001ac,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x200,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001200],   # Enable double buffering
    'min_program_length' : 0x200,

    # Flash information
    'flash_start': 0x0,
    'flash_size': 0x20000,
    'sector_sizes': (
        (0x0, 0x200),
    )
}


class HC32M423xA(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x20000, page_size=0x200, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x1FFFE000, length=0x4000)
        )

    def __init__(self, session):
        super(HC32M423xA, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32M423.svd")

    def post_connect_hook(self):
        self.write32(DBGMCU.STPCTL, DBGMCU.STPCTL_VALUE)
        self.write32(DBGMCU.TRACECTL, DBGMCU.TRACECTL_VALUE)



# pyOCD debugger
# Copyright (c) 2019 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Callable, List, Optional, Sequence, Type, TypeVar, Union, cast)

_T = TypeVar("_T", bound="GraphNode")

class GraphNode:
    """@brief Simple graph node.

    All nodes have a parent, which is None for a root node, and zero or more children.
    Nodes optionally have a name.

    Supports indexing and iteration over children.
    """

    def __init__(self) -> None:
        """@brief Constructor."""
        super().__init__()
        self._parent: Optional[GraphNode] = None
        self._children: List[GraphNode] = []
        self._node_name: Optional[str] = None

    @property
    def node_name(self) -> Optional[str]:
        """@brief Name of this graph node."""
        return self._node_name

    @node_name.setter
    def node_name(self, new_name: str) -> None:
        self._node_name = new_name

    @property
    def parent(self) -> Optional["GraphNode"]:
        """@brief This node's parent in the object graph."""
        return self._parent

    @property
    def children(self) -> Sequence["GraphNode"]:
        """@brief Child nodes in the object graph."""
        return self._children

    @property
    def is_leaf(self) -> bool:
        """@brief Returns true if the node has no children."""
        return len(self.children) == 0

    def add_child(self, node: "GraphNode") -> None:
        """@brief Link a child node onto this object."""
        node._parent = self
        self._children.append(node)

    def find_root(self) -> "GraphNode":
        """@brief Returns the root node of the object graph."""
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    def find_children(self,
            predicate: Callable[["GraphNode"], bool],
            breadth_first: bool = True
        ) -> Sequence["GraphNode"]:
        """@brief Recursively search for children that match a given predicate.
        @param self
        @param predicate A callable accepting a single argument for the node to examine. If the
            predicate returns True, then that node is added to the result list and no further
            searches on that node's children are performed. A False predicate result causes the
            node's children to be searched.
        @param breadth_first Whether to search breadth first. Pass False to search depth first.
        @returns List of matching child nodes, or an empty list if no matches were found.
        """
        def _search(node: GraphNode):
            results: List[GraphNode] = []
            childrenToExamine: List[GraphNode] = []
            for child in node.children:
                if predicate(child):
                    results.append(child)
                elif not breadth_first:
                    results.extend(_search(child))
                elif breadth_first:
                    childrenToExamine.append(child)

            if breadth_first:
                for child in childrenToExamine:
                    results.extend(_search(child))
            return results

        return _search(self)

    def get_first_child_of_type(self, klass: Type[_T]) -> Optional[_T]:
        """@brief Breadth-first search for a child of the given class.
        @param self
        @param klass The class type to search for. The first child at any depth that is an instance
            of this class or a subclass thereof will be returned. Matching children at more shallow
            nodes will take precedence over deeper nodes.
        @returns Either a node object or None.
        """
        matches = self.find_children(lambda c: isinstance(c, klass))
        if len(matches):
            return cast(Optional[_T], matches[0])
        else:
            return None

    def __getitem__(self, key: Union[int, str, slice]) -> Union["GraphNode", List["GraphNode"]]:
        """@brief Returns the child with the given index or node name.

        Slicing is supported with integer indexes.
        """
        if isinstance(key, str):
            # Replace with dict at some point.
            for c in self._children:
                if c.node_name == key:
                    return c
            else:
                raise KeyError(f"no child node with name '{key}'")
        else:
            return self._children[key]

    def __iter__(self):
        """@brief Iterate over the node's children."""
        return iter(self.children)

    def _dump_desc(self) -> str:
        """@brief Similar to __repr__ by used for dump_to_str()."""
        return str(self)

    def dump_to_str(self) -> str:
        """@brief Returns a string describing the object graph."""

        def _dump(node, level):
            result = ("  " * level) + "- " + node._dump_desc() + "\n"
            for child in node.children:
                result += _dump(child, level + 1)
            return result

        return _dump(self, 0)

    def dump(self) -> None:
        """@brief Pretty print the object graph to stdout."""
        print(self.dump_to_str())



# pyOCD debugger
# Copyright (c) 2006-2013,2018 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# Copyright (c) 2023 Benjamin Sølberg
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (Any, Optional, TYPE_CHECKING)

from ..core import exceptions
from ..target import (TARGET, normalise_target_type_name)
from ..target.pack import pack_target
from ..utility.graph import GraphNode

if TYPE_CHECKING:
    from ..core.session import Session
    from .board_ids import BoardInfo

LOG = logging.getLogger(__name__)

class Board(GraphNode):
    """@brief Represents the board containing the target and associated components.

    The board is the root of the runtime object graph. Responsible for creating the Target instance
    corresponding to the indicated target type name.
    """
    def __init__(self,
            session: "Session",
            target: Optional[str] = None,
            board_info: Optional["BoardInfo"] = None,
            ) -> None:
        """@brief Constructor

        This method is responsible for selecting the SoCTarget subclass for the SoC, implementing the target
        type support. There are several possible sources for the target type name, with differing levels of
        priority.

        1. `target` parameter
        2. `target_override` session option
        3. `board_info.target` parameter
        4. Last resort `cortex_m` target type. If this type is used, a warning is printed (unless the
            `warning.cortex_m_default` option is disabled).

        @param self
        @param session The session instance that owns us.
        @param target Target type name to use. If this parameter is set, it overrides all other sources of the
            target type.
        @param board_info A `BoardInfo` object with various descriptive information about the board. The
            `target` attribute is used as a secondary target type name, with lower precedence than the
            `target_override` session option.
        """
        super().__init__()

        # Use the session option if no target type was given to us.
        if target is None:
            if session.options.is_set('target_override'):
                target = session.options.get('target_override')
            elif board_info:
                target = board_info.target

        # As a last resort, default the target to 'cortex_m'.
        if target is None:
            target = 'cortex_m'

            # Log a helpful warning when defaulting to the generic cortex_m target.
            if session.options.get('warning.cortex_m_default'):
                LOG.warning("Generic 'cortex_m' target type is selected by default; is this "
                            "intentional? You will be able to debug most devices, but not program "
                            "flash. To set the target type use the '--target' argument or "
                            "'target_override' option. Use 'pyocd list --targets' to see available "
                            "targets types.")

        assert target is not None

        # Convert dashes to underscores in the target type, and convert to lower case.
        target = normalise_target_type_name(target)

        # Write the effective target type back to options if it's different.
        if target != session.options.get('target_override'):
            session.options['target_override'] = target

        self._session = session
        self._target_type = target
        self._info = board_info
        self._test_binary = board_info.binary if (board_info and board_info.binary) \
                else session.options.get('test_binary')
        self._delegate = None
        self._inited = False

        # Create targets from provided CMSIS pack.
        if session.options['pack'] is not None:
            pack_target.PackTargets.populate_targets_from_pack(session.options['pack'])

        # Create targets from the cmsis-pack-manager cache.
        if self._target_type not in TARGET:
            pack_target.ManagedPacks.populate_target(target)

        # Create Target instance.
        try:
            self.target = TARGET[self._target_type](session)
        except KeyError as exc:
            raise exceptions.TargetSupportError(
                f"Target type {self._target_type} not recognized. Use 'pyocd list --targets' to see currently "
                "available target types. "
                "See <https://pyocd.io/docs/target_support.html> "
                "for how to install additional target support.") from exc

        # Tell the user what target type is selected.
        LOG.info("Target type is %s", self._target_type)

        self._name = board_info.name if (board_info and board_info.name) \
                else f"Generic {self.target_type} board"
        self._vendor = board_info.vendor if (board_info and board_info.vendor) else ""

        # Standard graph node name.
        self.node_name = 'board'

        self.add_child(self.target)

    def init(self) -> None:
        """@brief Initialize the board."""
        # If we don't have a delegate set yet, see if there is a session delegate.
        if (self.delegate is None) and (self.session.delegate is not None):
            self.delegate = self.session.delegate

        # Delegate pre-init hook.
        if (self.delegate is not None) and hasattr(self.delegate, 'will_connect'):
            self.delegate.will_connect(board=self)

        # Init the target.
        self.target.init()
        self._inited = True

        # Delegate post-init hook.
        if (self.delegate is not None) and hasattr(self.delegate, 'did_connect'):
            self.delegate.did_connect(board=self)

    def uninit(self) -> None:
        """@brief Uninitialize the board."""
        if self._inited:
            LOG.debug("uninit board %s", self)
            resume = self.session.options.get('resume_on_disconnect')
            self.target.disconnect(resume)
            self._inited = False

    @property
    def session(self) -> "Session":
        """@brief The session that owns this board instance."""
        return self._session

    @property
    def delegate(self) -> Any:
        """@brief Delegate object that will be inherited by the SoCTarget."""
        return self._delegate

    @delegate.setter
    def delegate(self, the_delegate: Any) -> None:
        """@brief Set the delegate object that will be inherited by the SoCTarget."""
        self._delegate = the_delegate

    @property
    def unique_id(self) -> str:
        """@brief The probe's unique ID.

        Deprecated. Use the probe's `unique_id` property instead.
        """
        assert self.session.probe
        return self.session.probe.unique_id

    @property
    def target_type(self) -> str:
        """@brief Target type name."""
        return self._target_type

    @property
    def test_binary(self) -> Optional[str]:
        return self._test_binary

    @property
    def vendor(self) -> str:
        """@brief The board's vendor name."""
        return self._vendor

    @property
    def name(self) -> str:
        """@brief The board's name."""
        return self._name

    # Deprecated property.
    @property
    def description(self) -> str:
        """@brief Return description of the board."""
        return self.name



# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2019 Monadnock Systems Ltd.
# Copyright (c) 2023 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from time import sleep
import os
import re
from zipfile import ZipFile
from tempfile import TemporaryDirectory
from intelhex import IntelHex

from ...core import exceptions
from ...core.memory_map import FlashRegion
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...flash.eraser import FlashEraser
from ...flash.file_programmer import FileProgrammer
from ...utility.timeout import Timeout
from ...utility.progress import print_progress
from ...commands.base import CommandBase
from ...commands.execution_context import CommandSet

from typing import (Callable, Optional, TYPE_CHECKING, Union)
ProgressCallback = Callable[[Union[int, float]], None]

if TYPE_CHECKING:
    from ...core.session import Session

AHB_AP_NUM = 0x0
APB_AP_NUM = 0x3
CTRL_AP_NUM = 0x4

CTRL_AP_RESET = 0x000
CTRL_AP_ERASEALL = 0x004
CTRL_AP_ERASEALLSTATUS = 0x008
CTRL_AP_APPROTECTSTATUS = 0x00C
CTRL_AP_ERASEPROTECTSTATUS = 0x018
CTRL_AP_ERASEPROTECTDISABLE = 0x01C
CTRL_AP_MAILBOX_TXDATA = 0x020
CTRL_AP_MAILBOX_TXSTATUS = 0x024
CTRL_AP_MAILBOX_RXDATA = 0x028
CTRL_AP_MAILBOX_RXSTATUS = 0x02C
CTRL_AP_IDR = 0x0FC

CTRL_AP_ERASEALLSTATUS_READY = 0x0
CTRL_AP_ERASEALLSTATUS_BUSY = 0x1

CTRL_AP_APPROTECTSTATUS_APPROTECT_MSK = 0x1
CTRL_AP_APPROTECTSTATUS_SECUREAPPROTECT_MSK = 0x2
CTRL_AP_ERASEPROTECTSTATUS_MSK = 0x1

CTRL_AP_MAILBOX_STATUS_NODATAPENDING = 0x0
CTRL_AP_MAILBOX_STATUS_DATAPENDING = 0x1

CTRL_AP_RESET_NORESET = 0x0
CTRL_AP_RESET_RESET = 0x1

CTRL_AP_ERASEALL_NOOPERATION = 0x0
CTRL_AP_ERASEALL_ERASE = 0x1

CTRL_IDR_EXPECTED = 0x12880000

MASS_ERASE_TIMEOUT = 30.0

FAULT_EVENT = 0x4002A100
COMMAND_EVENT = 0x4002A108
DATA_EVENT = 0x4002A110
IPC_PIPELINED_MAX_BUFFER_SIZE = 0xE000
IPC_MAX_BUFFER_SIZE = 0x10000

CSW_DEVICEEN =  0x00000040

# just an empty program compiled with the Nordic MDK without ENABLE_APPROTECT set
nrf91_empty_image = [
    0x00, 0xe0, 0x03, 0x20, 0xe1, 0x04, 0x00, 0x00, 0x09, 0x05, 0x00, 0x00, 0x0b, 0x05, 0x00, 0x00, \
    0x0d, 0x05, 0x00, 0x00, 0x0f, 0x05, 0x00, 0x00, 0x11, 0x05, 0x00, 0x00, 0x13, 0x05, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, 0x05, 0x00, 0x00, \
    0x17, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x19, 0x05, 0x00, 0x00, 0x1b, 0x05, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x1d, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x10, 0xb5, 0x05, 0x4c, 0x23, 0x78, 0x33, 0xb9, 0x04, 0x4b, 0x13, 0xb1, \
    0x04, 0x48, 0x00, 0xe0, 0x00, 0xbf, 0x01, 0x23, 0x23, 0x70, 0x10, 0xbd, 0x3c, 0x04, 0x00, 0x20, \
    0x00, 0x00, 0x00, 0x00, 0x44, 0x07, 0x00, 0x00, 0x08, 0xb5, 0x03, 0x4b, 0x1b, 0xb1, 0x03, 0x49, \
    0x03, 0x48, 0x00, 0xe0, 0x00, 0xbf, 0x08, 0xbd, 0x00, 0x00, 0x00, 0x00, 0x40, 0x04, 0x00, 0x20, \
    0x44, 0x07, 0x00, 0x00, 0xa3, 0xf5, 0x80, 0x3a, 0x70, 0x47, 0x00, 0xbf, 0x17, 0x4b, 0x00, 0x2b, \
    0x08, 0xbf, 0x13, 0x4b, 0x9d, 0x46, 0xff, 0xf7, 0xf5, 0xff, 0x00, 0x21, 0x8b, 0x46, 0x0f, 0x46, \
    0x13, 0x48, 0x14, 0x4a, 0x12, 0x1a, 0x00, 0xf0, 0x5f, 0xf8, 0x0e, 0x4b, 0x00, 0x2b, 0x00, 0xd0, \
    0x98, 0x47, 0x0d, 0x4b, 0x00, 0x2b, 0x00, 0xd0, 0x98, 0x47, 0x00, 0x20, 0x00, 0x21, 0x04, 0x00, \
    0x0d, 0x00, 0x0d, 0x48, 0x00, 0x28, 0x02, 0xd0, 0x0c, 0x48, 0x00, 0xf0, 0x0d, 0xf9, 0x00, 0xf0, \
    0x27, 0xf8, 0x20, 0x00, 0x29, 0x00, 0x00, 0xf0, 0x71, 0xf9, 0x00, 0xf0, 0x11, 0xf8, 0x00, 0xbf, \
    0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0x03, 0x20, \
    0x3c, 0x04, 0x00, 0x20, 0x5c, 0x04, 0x00, 0x20, 0xf9, 0x03, 0x00, 0x00, 0x05, 0x04, 0x00, 0x00, \
    0x08, 0xb5, 0x00, 0x21, 0x04, 0x46, 0x00, 0xf0, 0x8b, 0xf8, 0x04, 0x4b, 0x18, 0x68, 0xc3, 0x6b, \
    0x03, 0xb1, 0x98, 0x47, 0x20, 0x46, 0x00, 0xf0, 0x4f, 0xf9, 0x00, 0xbf, 0x40, 0x07, 0x00, 0x00, \
    0x70, 0xb5, 0x0d, 0x4e, 0x0d, 0x4d, 0x76, 0x1b, 0xb6, 0x10, 0x06, 0xd0, 0x00, 0x24, 0x01, 0x34, \
    0x55, 0xf8, 0x04, 0x3b, 0x98, 0x47, 0xa6, 0x42, 0xf9, 0xd1, 0x09, 0x4e, 0x00, 0xf0, 0x6c, 0xfa, \
    0x08, 0x4d, 0x76, 0x1b, 0xb6, 0x10, 0x06, 0xd0, 0x00, 0x24, 0x01, 0x34, 0x55, 0xf8, 0x04, 0x3b, \
    0x98, 0x47, 0xa6, 0x42, 0xf9, 0xd1, 0x70, 0xbd, 0x30, 0x04, 0x00, 0x20, 0x30, 0x04, 0x00, 0x20, \
    0x38, 0x04, 0x00, 0x20, 0x30, 0x04, 0x00, 0x20, 0x70, 0xb5, 0x86, 0x07, 0x48, 0xd0, 0x54, 0x1e, \
    0x00, 0x2a, 0x3f, 0xd0, 0xca, 0xb2, 0x03, 0x46, 0x01, 0xe0, 0x01, 0x3c, 0x3a, 0xd3, 0x03, 0xf8, \
    0x01, 0x2b, 0x9d, 0x07, 0xf9, 0xd1, 0x03, 0x2c, 0x2d, 0xd9, 0xcd, 0xb2, 0x0f, 0x2c, 0x45, 0xea, \
    0x05, 0x25, 0x45, 0xea, 0x05, 0x45, 0x36, 0xd9, 0xa4, 0xf1, 0x10, 0x0c, 0x03, 0xf1, 0x20, 0x0e, \
    0x03, 0xf1, 0x10, 0x02, 0x2c, 0xf0, 0x0f, 0x06, 0x4f, 0xea, 0x1c, 0x1c, 0xb6, 0x44, 0x42, 0xe9, \
    0x04, 0x55, 0x42, 0xe9, 0x02, 0x55, 0x10, 0x32, 0x72, 0x45, 0xf8, 0xd1, 0x0c, 0xf1, 0x01, 0x02, \
    0x14, 0xf0, 0x0c, 0x0f, 0x04, 0xf0, 0x0f, 0x0c, 0x03, 0xeb, 0x02, 0x12, 0x13, 0xd0, 0xac, 0xf1, \
    0x04, 0x03, 0x23, 0xf0, 0x03, 0x03, 0x04, 0x33, 0x13, 0x44, 0x42, 0xf8, 0x04, 0x5b, 0x93, 0x42, \
    0xfb, 0xd1, 0x0c, 0xf0, 0x03, 0x04, 0x2c, 0xb1, 0xca, 0xb2, 0x1c, 0x44, 0x03, 0xf8, 0x01, 0x2b, \
    0x9c, 0x42, 0xfb, 0xd1, 0x70, 0xbd, 0x64, 0x46, 0x13, 0x46, 0x00, 0x2c, 0xf4, 0xd1, 0xf9, 0xe7, \
    0x03, 0x46, 0x14, 0x46, 0xbf, 0xe7, 0x1a, 0x46, 0xa4, 0x46, 0xe0, 0xe7, 0x02, 0x4b, 0x13, 0xb1, \
    0x02, 0x48, 0x00, 0xf0, 0x69, 0xb8, 0x70, 0x47, 0x00, 0x00, 0x00, 0x00, 0x05, 0x04, 0x00, 0x00, \
    0x2d, 0xe9, 0xf8, 0x4f, 0x2e, 0x4b, 0x83, 0x46, 0x0f, 0x46, 0x18, 0x68, 0x2d, 0xed, 0x02, 0x8b, \
    0x00, 0xf0, 0x76, 0xf8, 0x2b, 0x4b, 0xd3, 0xf8, 0x00, 0x80, 0xd8, 0xf8, 0x48, 0x61, 0xa6, 0xb1, \
    0x08, 0xee, 0x10, 0xba, 0x4f, 0xf0, 0x01, 0x09, 0x4f, 0xf0, 0x00, 0x0a, 0x74, 0x68, 0x65, 0x1e, \
    0x0b, 0xd4, 0x01, 0x34, 0x06, 0xeb, 0x84, 0x04, 0x7f, 0xb1, 0xd4, 0xf8, 0x00, 0x31, 0xbb, 0x42, \
    0x0b, 0xd0, 0x01, 0x3d, 0x04, 0x3c, 0x6b, 0x1c, 0xf6, 0xd1, 0x1d, 0x4b, 0xbd, 0xec, 0x02, 0x8b, \
    0x18, 0x68, 0xbd, 0xe8, 0xf8, 0x4f, 0x00, 0xf0, 0x55, 0xb8, 0x73, 0x68, 0x22, 0x68, 0x01, 0x3b, \
    0xab, 0x42, 0x0c, 0xbf, 0x75, 0x60, 0xc4, 0xf8, 0x00, 0xa0, 0x00, 0x2a, 0xe9, 0xd0, 0x09, 0xfa, \
    0x05, 0xf1, 0xd6, 0xf8, 0x88, 0x01, 0xd6, 0xf8, 0x04, 0xb0, 0x01, 0x42, 0x0b, 0xd1, 0x90, 0x47, \
    0x72, 0x68, 0x5a, 0x45, 0x13, 0xd1, 0xd8, 0xf8, 0x48, 0x31, 0xb3, 0x42, 0xd9, 0xd0, 0x1e, 0x46, \
    0x00, 0x2e, 0xcb, 0xd1, 0xd9, 0xe7, 0xd6, 0xf8, 0x8c, 0x01, 0x01, 0x42, 0x0c, 0xd1, 0xd4, 0xf8, \
    0x80, 0x10, 0x18, 0xee, 0x10, 0x0a, 0x90, 0x47, 0x72, 0x68, 0x5a, 0x45, 0xeb, 0xd0, 0xd8, 0xf8, \
    0x48, 0x61, 0x00, 0x2e, 0xba, 0xd1, 0xc8, 0xe7, 0xd4, 0xf8, 0x80, 0x00, 0x90, 0x47, 0xdf, 0xe7, \
    0x28, 0x04, 0x00, 0x20, 0x40, 0x07, 0x00, 0x00, 0x00, 0x23, 0x01, 0x46, 0x1a, 0x46, 0x18, 0x46, \
    0x00, 0xf0, 0x1a, 0xb8, 0x38, 0xb5, 0x08, 0x4b, 0x08, 0x4d, 0x5b, 0x1b, 0x9c, 0x10, 0x07, 0xd0, \
    0x04, 0x3b, 0x1d, 0x44, 0x01, 0x3c, 0x55, 0xf8, 0x04, 0x39, 0x98, 0x47, 0x00, 0x2c, 0xf9, 0xd1, \
    0xbd, 0xe8, 0x38, 0x40, 0x00, 0xf0, 0x86, 0xb9, 0x3c, 0x04, 0x00, 0x20, 0x38, 0x04, 0x00, 0x20, \
    0x70, 0x47, 0x00, 0xbf, 0x70, 0x47, 0x00, 0xbf, 0x2d, 0xe9, 0xf0, 0x47, 0xdf, 0xf8, 0x84, 0xa0, \
    0x06, 0x46, 0x98, 0x46, 0x0f, 0x46, 0xda, 0xf8, 0x00, 0x00, 0x91, 0x46, 0xff, 0xf7, 0xf0, 0xff, \
    0x1b, 0x4b, 0x1b, 0x68, 0xd3, 0xf8, 0x48, 0x41, 0x34, 0xb3, 0x65, 0x68, 0x1f, 0x2d, 0x28, 0xdc, \
    0x5e, 0xb9, 0x6b, 0x1c, 0x02, 0x35, 0xda, 0xf8, 0x00, 0x00, 0x63, 0x60, 0x44, 0xf8, 0x25, 0x70, \
    0xff, 0xf7, 0xe0, 0xff, 0x00, 0x20, 0xbd, 0xe8, 0xf0, 0x87, 0x04, 0xeb, 0x85, 0x01, 0x01, 0x23, \
    0x02, 0x2e, 0xc1, 0xf8, 0x88, 0x90, 0x03, 0xfa, 0x05, 0xf3, 0xd4, 0xf8, 0x88, 0x21, 0x42, 0xea, \
    0x03, 0x02, 0xc4, 0xf8, 0x88, 0x21, 0xc1, 0xf8, 0x08, 0x81, 0xe2, 0xd1, 0xd4, 0xf8, 0x8c, 0x21, \
    0x13, 0x43, 0xc4, 0xf8, 0x8c, 0x31, 0xdc, 0xe7, 0x03, 0xf5, 0xa6, 0x74, 0xc3, 0xf8, 0x48, 0x41, \
    0xd3, 0xe7, 0xda, 0xf8, 0x00, 0x00, 0xff, 0xf7, 0xbd, 0xff, 0x4f, 0xf0, 0xff, 0x30, 0xda, 0xe7, \
    0x40, 0x07, 0x00, 0x00, 0x28, 0x04, 0x00, 0x20, 0xfe, 0xe7, 0x00, 0xbf, 0x5a, 0x22, 0x03, 0x4b, \
    0xc3, 0xf8, 0x10, 0x2e, 0xc3, 0xf8, 0x00, 0x2e, 0xfe, 0xe7, 0x00, 0xbf, 0x00, 0x90, 0x03, 0x50, \
    0x06, 0x49, 0x07, 0x4a, 0x07, 0x4b, 0x9b, 0x1a, 0x03, 0xdd, 0x04, 0x3b, 0xc8, 0x58, 0xd0, 0x50, \
    0xfb, 0xdc, 0x00, 0xf0, 0x15, 0xf8, 0xff, 0xf7, 0x51, 0xfe, 0x00, 0x00, 0x68, 0x07, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x20, 0x3c, 0x04, 0x00, 0x20, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, \
    0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0xfe, 0xe7, 0x00, 0xbf, \
    0x4f, 0xf4, 0x7f, 0x03, 0xd3, 0xf8, 0x30, 0x21, 0xd3, 0xf8, 0x34, 0x31, 0xbf, 0xf3, 0x4f, 0x8f, \
    0x09, 0x2a, 0x06, 0xd1, 0x4f, 0xf0, 0x50, 0x23, 0x00, 0x22, 0xc3, 0xf8, 0x14, 0x21, 0xc3, 0xf8, \
    0x18, 0x21, 0x4f, 0xf4, 0x7f, 0x03, 0xd3, 0xf8, 0x30, 0x21, 0xd3, 0xf8, 0x34, 0x31, 0xbf, 0xf3, \
    0x4f, 0x8f, 0x09, 0x2a, 0x7e, 0xd0, 0x4f, 0xf4, 0x7f, 0x03, 0xd3, 0xf8, 0x30, 0x21, 0xd3, 0xf8, \
    0x34, 0x31, 0xbf, 0xf3, 0x4f, 0x8f, 0x09, 0x2a, 0x6d, 0xd0, 0x4f, 0xf4, 0x7f, 0x03, 0xd3, 0xf8, \
    0x30, 0x21, 0xd3, 0xf8, 0x34, 0x31, 0xbf, 0xf3, 0x4f, 0x8f, 0x09, 0x2a, 0x5c, 0xd0, 0x4f, 0xf4, \
    0x7f, 0x03, 0xd3, 0xf8, 0x30, 0x21, 0xd3, 0xf8, 0x34, 0x31, 0xbf, 0xf3, 0x4f, 0x8f, 0x09, 0x2a, \
    0x06, 0xd1, 0x00, 0x21, 0x01, 0x22, 0x58, 0x4b, 0xc3, 0xf8, 0x0c, 0x17, 0xc3, 0xf8, 0x10, 0x27, \
    0x56, 0x4b, 0x57, 0x48, 0x19, 0x46, 0x03, 0xf1, 0xf8, 0x52, 0x09, 0x68, 0x04, 0x3b, 0x02, 0xf5, \
    0x9c, 0x22, 0x83, 0x42, 0x11, 0x60, 0xf5, 0xd1, 0xbf, 0xf3, 0x4f, 0x8f, 0x51, 0x48, 0x00, 0x22, \
    0xd0, 0xf8, 0x7c, 0x37, 0x50, 0x49, 0x23, 0xf0, 0x10, 0x03, 0xc0, 0xf8, 0x7c, 0x37, 0x07, 0xe0, \
    0xd3, 0xf8, 0x00, 0x03, 0xd3, 0xf8, 0x04, 0x33, 0xb2, 0xf5, 0x80, 0x7f, 0x03, 0x60, 0x0b, 0xd0, \
    0x08, 0x68, 0xbf, 0xf3, 0x4f, 0x8f, 0xd3, 0x00, 0x01, 0x30, 0x03, 0xf5, 0x7f, 0x03, 0x02, 0xf1, \
    0x01, 0x02, 0x01, 0xf1, 0x08, 0x01, 0xeb, 0xd1, 0x44, 0x4b, 0xda, 0x69, 0xbf, 0xf3, 0x4f, 0x8f, \
    0xd1, 0x07, 0x30, 0xd4, 0x1b, 0x6a, 0xbf, 0xf3, 0x4f, 0x8f, 0x01, 0x33, 0x2b, 0xd0, 0x40, 0x4b, \
    0xd3, 0xf8, 0x8c, 0x20, 0x42, 0xf4, 0x40, 0x62, 0xc3, 0xf8, 0x8c, 0x20, 0xd3, 0xf8, 0x88, 0x20, \
    0x42, 0xf4, 0x70, 0x02, 0xc3, 0xf8, 0x88, 0x20, 0xbf, 0xf3, 0x4f, 0x8f, 0xbf, 0xf3, 0x6f, 0x8f, \
    0x38, 0x4b, 0x39, 0x4a, 0x1a, 0x60, 0x70, 0x47, 0x01, 0x2b, 0xa0, 0xd1, 0x0e, 0x22, 0x37, 0x4b, \
    0xc3, 0xf8, 0xe4, 0x2e, 0x9b, 0xe7, 0x01, 0x2b, 0x8f, 0xd0, 0x01, 0x22, 0x2a, 0x4b, 0xc3, 0xf8, \
    0x78, 0x25, 0x8a, 0xe7, 0x01, 0x2b, 0x7f, 0xf4, 0x7e, 0xaf, 0x27, 0x4a, 0xc2, 0xf8, 0x38, 0x3a, \
    0xc2, 0xf8, 0x78, 0x35, 0x77, 0xe7, 0xbf, 0xf3, 0x4f, 0x8f, 0x2d, 0x4a, 0xd2, 0xf8, 0x00, 0x34, \
    0x01, 0x2b, 0xfb, 0xd1, 0x2a, 0x49, 0xc2, 0xf8, 0x04, 0x35, 0xd1, 0xf8, 0x00, 0x34, 0x01, 0x2b, \
    0xfb, 0xd1, 0x22, 0x4b, 0xda, 0x69, 0xbf, 0xf3, 0x4f, 0x8f, 0xd2, 0x07, 0x0a, 0xd5, 0xda, 0x69, \
    0x22, 0xf0, 0x01, 0x02, 0xda, 0x61, 0xbf, 0xf3, 0x4f, 0x8f, 0x21, 0x4a, 0xd2, 0xf8, 0x00, 0x34, \
    0x01, 0x2b, 0xfb, 0xd1, 0x19, 0x4b, 0x1a, 0x6a, 0xbf, 0xf3, 0x4f, 0x8f, 0x01, 0x32, 0x14, 0xd0, \
    0x00, 0x23, 0x1b, 0x4a, 0xc2, 0xf8, 0x04, 0x35, 0xd2, 0xf8, 0x00, 0x34, 0x01, 0x2b, 0xfb, 0xd1, \
    0xbf, 0xf3, 0x4f, 0x8f, 0x12, 0x49, 0x17, 0x4b, 0xca, 0x68, 0x02, 0xf4, 0xe0, 0x62, 0x13, 0x43, \
    0xcb, 0x60, 0xbf, 0xf3, 0x4f, 0x8f, 0x00, 0xbf, 0xfd, 0xe7, 0x1a, 0x6a, 0x22, 0xf0, 0xff, 0x02, \
    0x42, 0xf0, 0x20, 0x02, 0x1a, 0x62, 0xbf, 0xf3, 0x4f, 0x8f, 0x0d, 0x4a, 0xd2, 0xf8, 0x00, 0x34, \
    0x01, 0x2b, 0xfb, 0xd1, 0xdc, 0xe7, 0x00, 0xbf, 0x00, 0x40, 0x00, 0x50, 0xfc, 0x0f, 0xff, 0x00, \
    0xfc, 0xff, 0xfe, 0x00, 0x00, 0x30, 0x00, 0x50, 0x00, 0x03, 0xff, 0x00, 0x00, 0x80, 0xff, 0x00, \
    0x00, 0xed, 0x00, 0xe0, 0x2c, 0x04, 0x00, 0x20, 0x00, 0x90, 0xd0, 0x03, 0x00, 0xa0, 0x03, 0x50, \
    0x00, 0x90, 0x03, 0x50, 0x04, 0x00, 0xfa, 0x05, 0xf8, 0xb5, 0x00, 0xbf, 0xf8, 0xbc, 0x08, 0xbc, \
    0x9e, 0x46, 0x70, 0x47, 0xf8, 0xb5, 0x00, 0xbf, 0xf8, 0xbc, 0x08, 0xbc, 0x9e, 0x46, 0x70, 0x47, \
    0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x34, 0xfa, 0xff, 0x7f, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xec, 0x02, 0x00, 0x20, \
    0x54, 0x03, 0x00, 0x20, 0xbc, 0x03, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0e, 0x33, 0xcd, 0xab, 0x34, 0x12, 0x6d, 0xe6, \
    0xec, 0xde, 0x05, 0x00, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x58, 0x04, 0x00, 0x20, 0x00, 0x90, 0xd0, 0x03, 0x1d, 0x03, 0x00, 0x00, 0x79, 0x01, 0x00, 0x00, \
    0x55, 0x01, 0x00, 0x00
]

LOG = logging.getLogger(__name__)


def change_endianness(x: int, n=4) -> int:
    return sum(((x >> 8*i) & 0xFF) << 8*(n-i-1) for i in range(n))

def bytes_to_word(bts):
    result = 0
    for i, b in enumerate(bts):
        result |= b << (8*i)
    return result

def word_to_bytes(wrd):
    result = []
    for i in range(4):
        result.append((wrd >> (8*i)) & 0xFF)
    return bytes(result)

def split_addr_range_into_chunks(range, chunk_size):
    chunks = []
    addr = range[0]
    while True:
        c = (addr, min(range[1], addr + chunk_size))
        chunks.append(c)
        addr = c[1]
        if addr == range[1]:
            break
    return chunks

FLASH_ALGO_UICR = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe7fdbe00,
    0xf240b570, 0xf2c00604, 0x25000600, 0x0006eb09, 0x5006f849, 0x5501e9c0, 0xb2d060c5, 0xf0004614,
    0x2800f9c9, 0xeb09bf1e, 0xe9c10106, 0x46050402, 0xbd704628, 0x4604b510, 0xf0002000, 0xf240f9bb,
    0x28000104, 0x0100f2c0, 0xeb09bf1c, 0xe9c20201, 0xeb090402, 0x68920201, 0xbf082a00, 0xf244bd10,
    0xf2c20200, 0x23010200, 0xf8596013, 0x44493001, 0x684b6053, 0x688b6093, 0x68c960d3, 0xbd106111,
    0xf240b580, 0xf2c00004, 0x21010000, 0x1000f849, 0x21004448, 0x1101e9c0, 0xf00060c1, 0x2000f9ad,
    0xbf00bd80, 0xf240b510, 0xf2c00004, 0x21020000, 0x1000f849, 0x24004448, 0x4401e9c0, 0xf00060c4,
    0x4601f93b, 0x0000f248, 0x00fff2c0, 0xf00022ff, 0xb138f80f, 0xf9a6f000, 0xd0052869, 0x28004604,
    0x2400bf08, 0xbd104620, 0x46202400, 0xbf00bd10, 0xf240b570, 0xf2c00c04, 0x23050c00, 0x300cf849,
    0x030ceb09, 0x0e00f04f, 0xe9c30784, 0xf8c3ee01, 0xd006e00c, 0x010ceb09, 0xe9c12203, 0x20652002,
    0xeb09bd70, 0xf04f030c, 0x078c0e01, 0xe004f8c3, 0xeb09d006, 0x2203000c, 0x2102e9c0, 0xbd702065,
    0x7efff648, 0x0efff2c0, 0x030ceb09, 0x45702402, 0xd906605c, 0x010ceb09, 0xe9c12204, 0x20662002,
    0x180bbd70, 0x0501f10e, 0x040ceb09, 0x42ab2603, 0xd9066066, 0x000ceb09, 0xe9c02104, 0x20661302,
    0x2900bd70, 0x2000bf04, 0xbf00bd70, 0x42937803, 0x3001d105, 0xf04f3901, 0xd1f70300, 0xeb09e005,
    0x2205010c, 0xe9c12301, 0x46182002, 0xbf00bd70, 0xf240b570, 0xf2c00304, 0xf04f0300, 0xeb090c03,
    0x24000503, 0xf849078e, 0xe9c5c003, 0x60ec4401, 0xeb09d005, 0xe9c00003, 0x2065c102, 0xf648bd70,
    0xf2c076ff, 0xeb0906ff, 0x24020503, 0x606c42b0, 0xeb09d906, 0x22040103, 0x2002e9c1, 0xbd702066,
    0x3601180d, 0x0403eb09, 0x0c03f04f, 0xf8c442b5, 0xd906c004, 0x0003eb09, 0xe9c02104, 0x20661502,
    0x2300bd70, 0x0f91ebb3, 0x088cd00b, 0x1f161f05, 0x0f04f856, 0x0f04f845, 0xf8d6f000, 0xd1f73c01,
    0x46182300, 0xbf00bd70, 0xf240b570, 0xf2c00c04, 0x23040c00, 0x300cf849, 0x030ceb09, 0x0e00f04f,
    0xe9c30784, 0xf8c3ee01, 0xd006e00c, 0x010ceb09, 0xe9c12203, 0x20652002, 0xeb09bd70, 0xf04f030c,
    0x078c0e01, 0xe004f8c3, 0xeb09d006, 0x2203000c, 0x2102e9c0, 0xbd702065, 0x7efff648, 0x0efff2c0,
    0x030ceb09, 0x45702402, 0xd906605c, 0x010ceb09, 0xe9c12204, 0x20662002, 0x180bbd70, 0x0501f10e,
    0x040ceb09, 0x42ab2603, 0xd9066066, 0x000ceb09, 0xe9c02104, 0x20661302, 0x2500bd70, 0x0f91ebb5,
    0x060ceb09, 0x0504f04f, 0xd0116075, 0xbf000889, 0x68156806, 0xd10542ae, 0x39013004, 0x0204f102,
    0xe005d1f6, 0x010ceb09, 0x46032206, 0x2002e9c1, 0xbd704618, 0x1030f240, 0x00fff2c0, 0x31016801,
    0x6800bf1c, 0xf6404770, 0xf2cf71e0, 0x78080100, 0xf3616849, 0x4770200b, 0x2020f240, 0x00fff2c0,
    0x31016801, 0x6800bf14, 0x5080f44f, 0xbf004770, 0x2024f240, 0x00fff2c0, 0x31016801, 0x6800bf14,
    0x7000f44f, 0xbf004770, 0x47702000, 0x47702000, 0xf7ffb510, 0x4604ffe1, 0xffeaf7ff, 0xf004fb00,
    0xbf00bd10, 0x42814401, 0x2001bf9c, 0xbf004770, 0x2b04f850, 0xbf1c3201, 0x47702000, 0xbf244288,
    0x47702001, 0xbf00e7f4, 0xbf004770, 0x47702000, 0x47702003, 0xbf842803, 0x47702069, 0xb240b580,
    0xf851a105, 0xf2490020, 0xf2c55104, 0x60080103, 0xf80af000, 0xbd802000, 0x00000000, 0x00000002,
    0x00000001, 0x00000000, 0x4000f249, 0x0003f2c5, 0x29006801, 0x4770d0fc, 0xf249b580, 0xf2c5500c,
    0x21010003, 0xf7ff6001, 0xbd80ffef, 0xb082b580, 0x98019001, 0x31fff04f, 0xf7ff6001, 0xb002ffe5,
    0xbf00bd80, 0x47702069, 0xf7ffb5b0, 0xb148ffa1, 0x25004604, 0xf7ff4628, 0xf7ffffe9, 0x4405ff7d,
    0xd3f742a5, 0xbdb02000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000005,
    'pc_unInit': 0x20000039,
    'pc_program_page': 0x200001b5,
    'pc_erase_sector': 0x200000a9,
    'pc_eraseAll': 0x20000085,

    'static_base' : 0x20000000 + 0x00000004 + 0x00000448,
    'begin_stack' : 0x20003460,
    'end_stack' : 0x20002460,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x1000,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        0x20000460,
        0x20001460
    ],
    'min_program_length' : 0x1000,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x448,
    'rw_start': 0x44c,
    'rw_size': 0x4,
    'zi_start': 0x450,
    'zi_size': 0x10,

    # Flash information
    'flash_start': 0xff8000,
    'flash_size': 0x1000,
    'sector_sizes': (
        (0x0, 0x1000),
    )
}

class nRF91ModemFirmwareUpdateCommand(CommandBase):
    INFO = {
            'names': ['nrf91-update-modem-fw'],
            'group': 'nrf91',
            'category': 'nrf91',
            'nargs': '*',
            'usage': "[-f] mfw_nrf91xx_x.x.x.zip",
            'help': "Update modem firmware for an nRF91 target.",
            'extra_help':
                "If -f is specified, modem firmware is written to the device, "
                "even if the correct version is already present."
            }
    
    file_path = ""
    needs_update = False

    def parse(self, args):
        if len(args) == 0 or not args[-1].endswith(".zip"):
            raise exceptions.CommandError(f"invalid argument")
        self.file_path = args[-1]
        for a in args:
            if a == '-f':
                self.needs_update = True


    def execute(self):
        if not self.needs_update:
            try:
                ModemUpdater(self.context.session).verify(self.file_path)
            except:
                self.needs_update = True
        if self.needs_update:
            ModemUpdater(self.context.session).program_and_verify(self.file_path)

class NRF91(CoreSightTarget):

    VENDOR = "Nordic Semiconductor"

    def __init__(self, session, memory_map=None):
        super(NRF91, self).__init__(session, memory_map)
        if memory_map.get_region_for_address(0x00ff8000) is None:
            LOG.debug("Adding UICR region")
            memory_map.add_region(
                FlashRegion(
                    start=0x00ff8000,
                    length=0x1000,
                    blocksize=0x1000,
                    is_testable=False,
                    is_erasable=False,
                    algo=FLASH_ALGO_UICR,
                )
            )
        self.ctrl_ap = None
        self.was_locked = False

    def create_init_sequence(self):
        seq = super(NRF91, self).create_init_sequence()

        # Must check whether security is enabled, and potentially auto-unlock, before
        # any init tasks that require system bus access.
        seq.wrap_task('discovery',
            lambda seq: seq.insert_before('find_components',
                              ('check_ctrl_ap_idr', self.check_ctrl_ap_idr),
                              ('check_flash_security', self.check_flash_security),
                          )
            )
        seq.wrap_task('discovery',
            lambda seq: seq.insert_after('create_cores',
                              ('persist_unlock', self.persist_unlock),
                          )
            )
        seq.insert_before('post_connect_hook',
                          ('check_part_info', self.check_part_info))

        return seq

    def check_ctrl_ap_idr(self):
        self.ctrl_ap = self.dp.aps[CTRL_AP_NUM]

        # Check CTRL-AP ID.
        if self.ctrl_ap.idr != CTRL_IDR_EXPECTED:
            LOG.error("%s: bad CTRL-AP IDR (is 0x%08x)", self.part_number, self.ctrl_ap.idr)

    def ap_is_enabled(self):
        csw = self.dp.read_ap(AHB_AP_NUM << 24)
        return csw & CSW_DEVICEEN
    
    def persist_unlock(self):
        if self.session.options.get('auto_unlock') and self.has_hardened_approtect:
            # Write Unprotected to UICR.APPROTECT
            self.write_uicr(0x00FF8000, 0x50FA50FA)

            # Write Unprotected to UICR.SECUREAPPROTECT
            self.write_uicr(0x00FF802C, 0x50FA50FA)

        if self.session.options.get('auto_unlock') and self.has_hardened_approtect and self.was_locked:
            # write unlock image
            self.write_flash(0, nrf91_empty_image)

            self.reset()

    def check_flash_security(self):
        """@brief Check security and unlock device.

        This init task determines whether the device is locked (APPROTECT enabled). If it is,
        and if auto unlock is enabled, then perform a mass erase to unlock the device.

        This init task runs *before* cores are created.
        """

        target_id = self.dp.read_dp(0x24)
        if target_id & 0xFFF != 0x289:
            LOG.error(f"This doesn't look like a Nordic Semiconductor device!")
        if target_id & 0xF0000 != 0x90000:
            LOG.error(f"This doesn't look like an nRF91 devcice!")
        revision = target_id >> 28
        self.has_hardened_approtect = revision > 2

        if self.is_locked():
            self.was_locked = True
            if self.session.options.get('auto_unlock'):
                LOG.warning("%s APPROTECT enabled: will try to unlock via mass erase", self.part_number)

                unlock_successful = False
                for _ in range(3):
                    # Do the mass erase.
                    if not self.mass_erase():
                        continue

                    if not self.has_hardened_approtect:
                        # pin reset is needed for older versions
                        self.session.probe.reset()
                        self.pre_connect()
                        self.dp.connect()

                    # Check if AP was enabled
                    if self.ap_is_enabled():
                        unlock_successful = True
                        break
                    else:
                        LOG.debug("unlock was not successful, retrying")

                if not unlock_successful:
                    raise exceptions.TargetError("unable to unlock device")
                self._discoverer._create_1_ap(AHB_AP_NUM)
                self._discoverer._create_1_ap(APB_AP_NUM)
            else:
                LOG.warning("%s APPROTECT enabled: not automatically unlocking", self.part_number)
        else:
            LOG.info("%s not in secure state", self.part_number)

    def is_locked(self):
        status = self.ctrl_ap.read_reg(CTRL_AP_APPROTECTSTATUS)
        return (status & CTRL_AP_APPROTECTSTATUS_APPROTECT_MSK == 0) \
            or (status & CTRL_AP_APPROTECTSTATUS_SECUREAPPROTECT_MSK == 0)

    def is_eraseprotected(self):
        status = self.ctrl_ap.read_reg(CTRL_AP_ERASEPROTECTSTATUS)
        return status & CTRL_AP_ERASEPROTECTSTATUS_MSK == 0

    def mass_erase(self):
        if self.is_eraseprotected():
            LOG.warning("ERASEPROTECT is enabled.")
            if self.is_locked():
                LOG.error("If the firmware supports unlocking with a known 32-bit key,")
                LOG.error("then this is the only way to recover the device.")
                return False
            else:
                LOG.warning("Performing a chip erase instead.")
                eraser = FlashEraser(self.session, FlashEraser.Mode.CHIP)
                eraser._log_chip_erase = False
                eraser.erase()
                return True

        # See Nordic Whitepaper nWP-027 for magic numbers and order of operations from the vendor
        self.ctrl_ap.write_reg(CTRL_AP_ERASEALL, CTRL_AP_ERASEALL_ERASE)
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                status = self.ctrl_ap.read_reg(CTRL_AP_ERASEALLSTATUS)
                if status == CTRL_AP_ERASEALLSTATUS_READY:
                    break
                sleep(0.5)
            else:
                # Timed out
                LOG.error("Mass erase timeout waiting for ERASEALLSTATUS")
                return False
        sleep(0.01)
        self.ctrl_ap.write_reg(CTRL_AP_RESET, CTRL_AP_RESET_RESET)
        self.ctrl_ap.write_reg(CTRL_AP_RESET, CTRL_AP_RESET_NORESET)
        sleep(0.2)
        return True

    def check_part_info(self):
        partno = self.read32(0x00FF0140)
        hwrevision = self.read32(0x00FF0144)
        variant = self.read32(0x00FF0148)

        LOG.info(f"This appears to be an nRF{partno:X} " +
                 f"{word_to_bytes(variant).decode('ASCII', errors='ignore')} " +
         	 f"{word_to_bytes(hwrevision).decode('ASCII', errors='ignore')}")

    def write_uicr(self, addr: int, value: int):
        current_value = self.read32(addr)
        if ((current_value & value) != value) and (current_value != 0xFFFFFFFF):
            raise exceptions.TargetError("cannot write UICR value, mass_erase needed")

        self.write32(0x50039504, 1)  # NVMC.CONFIG = WriteEnable
        self._wait_nvmc_ready()
        self.write32(addr, value)
        self._wait_nvmc_ready()
        self.write32(0x50039504, 0)  # NVMC.CONFIG = ReadOnly
        self._wait_nvmc_ready()

    def write_flash(self, addr: int, bytes):
        assert len(bytes) % 4 == 0, "only full words can be written"
        self.write32(0x50039504, 1)  # NVMC.CONFIG = WriteEnable
        self._wait_nvmc_ready()
        for off in range(0, len(bytes), 4):
            self.write32(addr+off, bytes_to_word(bytes[off:off+4]))
            self._wait_nvmc_readynext()
        self.write32(0x50039504, 0)  # NVMC.CONFIG = ReadOnly
        self._wait_nvmc_ready()

    def add_target_command_groups(self, command_set: CommandSet) -> None:
        command_set.add_command_group('nrf91')

    def _wait_nvmc_ready(self):
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                if self.read32(0x50039400) != 0x00000000:  # NVMC.READY != BUSY
                    break
            else:
                raise exceptions.TargetError("wait for NVMC timed out")
            
    def _wait_nvmc_readynext(self):
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                if self.read32(0x50039408) != 0x00000000:  # NVMC.READYNEXT != BUSY
                    break
            else:
                raise exceptions.TargetError("wait for NVMC timed out")

class ModemUpdater(object):
    """@brief Implements the nRF91 Modem Update procedure like described in nAN-41"""
    _target: "Target"
    _session: "Session"
    _progress: Optional[ProgressCallback]
    _total_data_size: int
    _progress_offset: float
    _current_progress_fraction: float
    _chunk_size: int
    _pipelined: bool
    _segments: list
    _firmware_update_digest: str

    def __init__(self,
                 session: "Session",
                 progress: Optional[ProgressCallback] = None,
                 ):
        self._session = session
        self._target = session.board.target
        self._total_data_size = 0
        self._pipelined = False
        self._segments = []
        self._firmware_update_digest = None

        if progress is not None:
            self._progress = progress
        elif session.options.get('hide_programming_progress'):
            self._progress = None
        else:
            self._progress = print_progress()

        self._reset_state()

    def program_and_verify(self, mfw_zip: str):
        """@brief Program and verify modem firmware from ZIP file."""
        self._setup_device()
        self._process_zip_file(mfw_zip)

        LOG.info("programming modem firmware..")
        self._total_data_size = sum(r[1]-r[0] for s in self._segments for r in s.segments())
        for s in self._segments:
            self._program_segment(s)
        self._progress(1.0)
        LOG.info("modem firmware programmed.")

        LOG.info("verifying modem firmware..")
        self._verify()
        LOG.info("modem firmware verified.")

    def verify(self, mfw_zip: str):
        """@brief Just verify modem firmware from ZIP file."""
        self._setup_device()
        self._process_zip_file(mfw_zip)

        LOG.info("verifying modem firmware..")
        self._verify()
        LOG.info("modem firmware verified.")

    def _setup_device(self):
        """@brief initialize device to use modem DFU"""
        # init UICR.HFXOSR if necessary
        if self._target.read32(0x00FF801C) == 0xFFFFFFFF:
            LOG.warning("UICR.HFXOSR is not set, setting it to 0x0E")
            self._target.write_uicr(addr=0x00FF801C, value=0x0000000E)

        # init UICR.HFXOCNT if necessary
        if self._target.read32(0x00FF8020) == 0xFFFFFFFF:
            LOG.warning("UICR.HFXOCNT is not set, setting it to 0x20")
            self._target.write_uicr(addr=0x00FF8020, value=0x00000020)

        self._target.reset_and_halt(reset_type=Target.ResetType.SW)

        # 1. configure IPC to be in non-secure mode
        self._target.write32(addr=0x500038A8, value=0x00000002)

        # 2. configure IPC HW for DFU
        self._target.write32(addr=0x4002A514, value=0x00000002)
        self._target.write32(addr=0x4002A51C, value=0x00000008)
        self._target.write32(addr=0x4002A610, value=0x21000000)
        self._target.write32(addr=0x4002A614, value=0x00000000)
        self._target.write32(addr=0x4002A590, value=0x00000001)
        self._target.write32(addr=0x4002A598, value=0x00000004)
        self._target.write32(addr=0x4002A5A0, value=0x00000010)

        # 3. configure RAM as non-secure
        for n in range(32):
            self._target.write32(addr=0x50003700+(n*4), value=0x00000007)

        # 4. allocate memory in RAM
        self._target.write32(addr=0x20000000, value=0x80010000)
        self._target.write32(addr=0x20000004, value=0x2100000C)
        self._target.write32(addr=0x20000008, value=0x0003FC00)

        # 5. reset the modem
        self._target.write32(addr=0x50005610, value=0)
        self._target.write32(addr=0x50005614, value=1)
        self._target.write32(addr=0x50005610, value=1)
        self._target.write32(addr=0x50005614, value=0)
        self._target.write32(addr=0x50005610, value=0)

    def _process_zip_file(self, mfw_zip: str):
        """@brief extract the mfw ZIP file and load DFU loader"""
        digest_id = self._read_key_digest()
        modem_firmware_loader = None

        with TemporaryDirectory() as tmpdir:
            with ZipFile(mfw_zip, 'r') as zip_ref:
                zip_ref.extractall(tmpdir)
            files = os.listdir(tmpdir)

            # find modem firmware loader
            for f in files:
                if f.startswith(f"{digest_id}.ipc_dfu.signed_") and f.endswith(".ihex"):
                    modem_firmware_loader = os.path.join(tmpdir, f)
                    m = re.match(r"\.ipc_dfu\.signed_(\d+)\.(\d+)\.(\d+)\.ihex", f[7:])
                    if m:
                        loader_version = tuple(int(x) for x in m.groups())
                        LOG.info("modem_firmware_loader version: {}.{}.{}".format(
                            *loader_version))
                        if loader_version > (1, 1, 2):
                            LOG.info("using pipelined method")
                            self._pipelined = True
                    break
            if not modem_firmware_loader:
                raise exceptions.TargetError(
                    f"No compatible loader {digest_id}.ipc_dfu.signed_x.x.x.ihex found.")

            # find modem firmware segments
            for f in files:
                m = re.match(r"firmware\.update\.image\.segments\.(\d+).hex", f)
                if m:
                    self._segments.append(
                        (m.group(1), os.path.join(tmpdir, f)))
            self._segments.sort()
            self._segments = [IntelHex(s[1]) for s in self._segments]

            if len(self._segments) == 0:
                raise exceptions.TargetError("No modem firmware segments found")

            # parse segment digests
            with open(os.path.join(tmpdir, "firmware.update.image.digest.txt"), "r") as f:
                for line in f:
                    m = re.match(r"SHA256 of all ranges in ascending address order:\s*(\w{64})",
                                 line)
                    if m:
                        self._firmware_update_digest = m.group(1)
            if not self._firmware_update_digest:
                raise exceptions.TargetError("no firmware digest found")

            LOG.info("loading modem firmware loader..")
            FileProgrammer(self._session).program(
                modem_firmware_loader, file_format='hex')
            self._target.write32(0x4002A004, 0x00000001)  # start IPC task
            self._wait_and_ack_events()
            LOG.info("modem_firmware_loader started.")

    def _read_key_digest(self) -> str:
        """@brief read first word of modem key digest for choosing a loader"""
        self._wait_and_ack_events()
        digest_data = change_endianness(self._target.read32(0x20000010))
        return (f"{digest_data:08X}")[:7]

    def _program_segment(self, segment: IntelHex):
        """@brief program contents of segment HEX file using DFU loader"""
        if self._pipelined:
            bufsz = IPC_PIPELINED_MAX_BUFFER_SIZE
        else:
            bufsz = IPC_MAX_BUFFER_SIZE

        chunks = []
        for s in segment.segments():
            chunks += split_addr_range_into_chunks(s, bufsz)

        if self._pipelined:
            self._write_chunk(segment, chunks[0], 0)

            for i, c in enumerate(chunks):
                self._commit_chunk(c, i % 2)

                # write next chunk while current one is processed
                if (i + 1) < len(chunks):
                    self._write_chunk(segment, chunks[i + 1], (i + 1) % 2)

                self._wait_and_ack_events()
        else:
            for i, c in enumerate(chunks):
                self._write_chunk(segment, c, 0)
                self._commit_chunk(c, 0)
                self._wait_and_ack_events()

    def _write_chunk(self, segment: IntelHex, chunk, bank):
        """@brief write a chunk of the current segment to RAM"""
        start = chunk[0]
        size = chunk[1]-chunk[0]
        if self._pipelined:
            ram_address = 0x2000001C + IPC_PIPELINED_MAX_BUFFER_SIZE * bank
        else:
            ram_address = 0x20000018

        data = list(segment.tobinarray(start=start, size=size))
        data_words = [bytes_to_word(data[i:i+4])
                      for i in range(0, len(data), 4)]
        self._target.write_memory_block32(ram_address, data_words)
        self._current_progress_fraction = size / float(self._total_data_size)
        self._progress_cb(1.0)
        self._progress_offset += self._current_progress_fraction

    def _commit_chunk(self, chunk, bank):
        """@brief signal DFU loader that chunk is ready to be programmed"""
        buffer_offset = bank * IPC_PIPELINED_MAX_BUFFER_SIZE
        self._target.write32(0x20000010, chunk[0])
        self._target.write32(0x20000014, chunk[1]-chunk[0])
        if self._pipelined:
            self._target.write32(0x20000018, buffer_offset)
        if self._pipelined:
            # command = PIPELINE_WRITE
            self._target.write32(0x2000000C, 0x9)
        else:
            # command = WRITE
            self._target.write32(0x2000000C, 0x3)
        # start IPC task
        self._target.write32(0x4002A004, 1)

    def _verify(self):
        """@brief verify programmed modem firmware"""
        ranges_to_verify = []
        for s in self._segments:
            for r in s.segments():
                if r[0] < 0x1000000:
                    ranges_to_verify.append(r)

        # write given start, size pairs and number of entries
        self._target.write32(0x20000010, len(ranges_to_verify))
        for i, (start, end) in enumerate(ranges_to_verify):
            self._target.write32(0x20000014 + (8 * i), start)
            self._target.write32(0x20000018 + (8 * i), end-start)

        # command = VERIFY
        self._target.write32(0x2000000C, 0x7)
        # start IPC task
        self._target.write32(0x4002A004, 1)

        self._wait_and_ack_events()

        response = self._target.read32(0x2000000C)
        if (response & 0xFF000000) == 0x5A000000:
            raise exceptions.TargetError(f"Error while verifying: {response & 0xFFFFFF:X}")

        digest_data = [self._target.read32(x) for x in range(0x20000010, 0x2000002D, 0x4)]
        digest_str = "".join(f"{x:08X}" for x in digest_data)

        if digest_str != self._firmware_update_digest:
            raise exceptions.TargetError(
                f"checksum mismatch: {digest_str} != {self._firmware_update_digest}"
            )

    def _wait_and_ack_events(self):
        """@brief wait for and acknowledge DFU events"""
        fault = False

        # poll for events
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                if self._target.read32(FAULT_EVENT) != 0:
                    fault = True
                    break
                if self._target.read32(COMMAND_EVENT) != 0:
                    break
                if self._target.read32(DATA_EVENT) != 0:
                    break
            else:
                raise exceptions.TargetError("wait for events timed out")

        # reset events
        for reg in [FAULT_EVENT, COMMAND_EVENT, DATA_EVENT]:
            self._target.write32(reg, 0)

        response = self._target.read32(0x2000000C)
        if (response & 0xFF000000) == 0xA5000000:
            LOG.debug(f"ACK response, code {response:08X}")
        elif (response & 0xFF000000) == 0x5A000000:
            raise exceptions.TargetError(f"NACK response, code {response:08X}")

        if fault:
            raise exceptions.TargetError("modem triggered FAULT_EVENT")

    def _reset_state(self):
        """@brief Clear all state variables. """
        self._total_data_size = 0
        self._progress_offset = 0.0
        self._current_progress_fraction = 0.0

    def _progress_cb(self, amount):
        """@brief callback for updating the progress bar"""
        if self._progress is not None:
            self._progress((amount * self._current_progress_fraction) + self._progress_offset)



# pyOCD debugger
# Copyright (c) 2023 Brian Pugh
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...coresight.cortex_m import CortexM
from ...coresight.minimal_mem_ap import MinimalMemAP as MiniAP


class DBGMCU:
    CR = 0xE00E1004
    CR_VALUE = 0x7 # DBG_STANDBY | DBG_STOP | DBG_SLEEP

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe7fdbe00,
    0x4770ba40, 0x4770bac0, 0x0030ea4f, 0x00004770, 0x8f4ff3bf, 0xb5104770, 0x48ea4603, 0x61604cea,
    0x48e9bf00, 0xf0006900, 0x28000001, 0x48e7d1f9, 0x60604ce5, 0x606048e6, 0x4ce648e2, 0xbf006020,
    0x1f0048e4, 0xf0006800, 0x28000001, 0x48dfd1f8, 0x3c104ce0, 0x48de6020, 0xf8c44cdb, 0x46200104,
    0x200069c0, 0x4601bd10, 0x68c048d7, 0x0001f040, 0x60d04ad5, 0x380848d7, 0xf0406800, 0xf8c20001,
    0x2000010c, 0xbf004770, 0x690048cf, 0x0001f000, 0xd1f92800, 0x49cc48cb, 0x46086148, 0xf02068c0,
    0x60c80001, 0x68c04608, 0x0008f040, 0x460860c8, 0xf04068c0, 0x60c80020, 0x48c3bf00, 0xf0006900,
    0x28000001, 0x48c0d1f9, 0xf02068c0, 0x49be0008, 0xbf0060c8, 0x1f0048bf, 0xf0006800, 0x28000001,
    0x48b8d1f8, 0x600849bb, 0xf8d048b7, 0xf020010c, 0x49b50001, 0x010cf8c1, 0xf8d04608, 0xf040010c,
    0xf8c10008, 0x4608010c, 0x010cf8d0, 0x0020f040, 0x390849b0, 0xbf006008, 0x1f0048ae, 0xf0006800,
    0x28000001, 0x48abd1f8, 0x68003808, 0x0008f020, 0xf8c149a5, 0x2000010c, 0xb5104770, 0xf3c14601,
    0xf1b13247, 0xd3366f00, 0x6f01f1b1, 0x489ed233, 0x4ba16940, 0x4b9c4318, 0xbf006158, 0x6900489a,
    0x0004f000, 0xd1f92800, 0x68c04897, 0x50fef420, 0x60d84b95, 0x68c04618, 0xea432304, 0x43181382,
    0x60d84b91, 0x68c04618, 0x0020f040, 0xbf0060d8, 0x6900488d, 0x0004f000, 0xd1f92800, 0x68c0488a,
    0x0004f020, 0x60d84b88, 0x69004618, 0x0001f000, 0x2001b3f0, 0x4887bd10, 0x4b876800, 0x4b824318,
    0x0114f8c3, 0x4883bf00, 0x68001f00, 0x0004f000, 0xd1f82800, 0x3808487f, 0xf4206800, 0x4b7a50fe,
    0x010cf8c3, 0x3808487b, 0xf1a26800, 0x24040380, 0x1383ea44, 0x4b744318, 0x010cf8c3, 0xf8d04618,
    0xf040010c, 0xf8c30020, 0xbf00010c, 0x1f004871, 0xf0006800, 0x28000004, 0x486ed1f8, 0x68003808,
    0x0004f020, 0xf8c34b68, 0x486a010c, 0xe0001f00, 0x6800e005, 0x0001f000, 0x2001b108, 0xf7ffe7ba,
    0x2000fee7, 0xb5f0e7b6, 0x46164603, 0x4635461a, 0xbf002400, 0x6900485c, 0x0001f000, 0xd1f92800,
    0x4f594858, 0xbf006178, 0x1f00485a, 0xf0006800, 0x28000001, 0x4853d1f8, 0x60384f56, 0x4852e09c,
    0xf02068c0, 0x4f500001, 0x463860f8, 0xf04068c0, 0x60f80002, 0x3808484f, 0xf0206800, 0xf8c70001,
    0x4638010c, 0x010cf8d0, 0x0002f040, 0x010cf8c7, 0xd30c2910, 0xe0062400, 0x6868682f, 0x60506017,
    0x32083508, 0x2c021c64, 0x3910dbf6, 0x2400e028, 0xf815e004, 0xf8020b01, 0x1c640b01, 0xd3f8428c,
    0xe0032400, 0xf80220ff, 0x1c640b01, 0x0010f1c1, 0xd8f742a0, 0x6f00f1b3, 0xf1b3d309, 0xd2066f01,
    0x68c04831, 0x0040f040, 0x60f84f2f, 0x4831e007, 0x68003808, 0x0040f040, 0xf8c74f2b, 0x2100010c,
    0xfe76f7ff, 0x6f00f1b3, 0xf1b3d30a, 0xd2076f01, 0x4825bf00, 0xf0006900, 0x28000001, 0xe007d1f9,
    0x4824bf00, 0x68001f00, 0x0001f000, 0xd1f82800, 0x6900481d, 0x4f1f2000, 0x683f1f3f, 0xb1b04300,
    0x6f00f1b3, 0xf1b3d309, 0xd2066f01, 0x68c04816, 0x0002f020, 0x60f84f14, 0x4816e007, 0x68003808,
    0x0002f020, 0xf8c74f10, 0x2001010c, 0xf1b3bdf0, 0xd3096f00, 0x6f01f1b3, 0x480bd206, 0xf02068c0,
    0x4f090002, 0xe00760f8, 0x3808480a, 0xf0206800, 0x4f050002, 0x010cf8c7, 0xf47f2900, 0x2000af60,
    0x0000e7e4, 0x0faf0000, 0x52002000, 0x45670123, 0xcdef89ab, 0x52002114, 0x0fef0000, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x2000001b,
    'pc_unInit': 0x2000006b,
    'pc_program_page': 0x2000024b,
    'pc_erase_sector': 0x2000013f,
    'pc_eraseAll': 0x2000008b,

    'static_base' : 0x20000000 + 0x00000004 + 0x000003dc,
    'begin_stack' : 0x200113f0,
    'end_stack' : 0x200103f0,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x8000,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        0x200003f0,
        0x200083f0
    ],
    'min_program_length' : 0x8000,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x3dc,
    'rw_start': 0x3e0,
    'rw_size': 0x4,
    'zi_start': 0x3e4,
    'zi_size': 0x0,

    # Flash information
    'flash_start': 0x8000000,
    'flash_size': 0x40000,
    'sector_sizes': (
        (0x0, 0x2000),
    )
}

class STM32H7B0xx(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        # Datasheet says there's 128KB, but there's actually 256KB
        FlashRegion( start=0x08000000, length=0x40000,  sector_size=0x2000,
                                                        page_size=0x8000,
                                                        is_boot_memory=True,
                                                        algo=FLASH_ALGO,
                                                        name="bank_1"),
        # Datasheet does not reference a flash bank 2,
        # but there's an additional 256KB here, as well.
        FlashRegion( start=0x08100000, length=0x40000,  sector_size=0x2000,
                                                        page_size=0x8000,
                                                        algo=FLASH_ALGO,
                                                        name="bank_2"),
        RamRegion(   start=0x20000000, length=0x20000, name="dtcm"),
        RamRegion(   start=0x24000000, length=0x40000, name="axi_sram_1"),
        RamRegion(   start=0x24040000, length=0x60000, name="axi_sram_2"),
        RamRegion(   start=0x240A0000, length=0x60000, name="axi_sram_3"),
        RamRegion(   start=0x30000000, length=0x10000, name="ahb_sram_1"),
        RamRegion(   start=0x30010000, length=0x10000, name="ahb_sram_2"),
        RamRegion(   start=0x38000000, length=0x8000, name="sdr_sram"),
        RamRegion(   start=0x38800000, length=0x1000, name="backup_sram"),
        )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)

    def assert_reset_for_connect(self):
        self.dp.assert_reset(1)

    def safe_reset_and_halt(self):
        assert self.dp.is_reset_asserted()

        # At this point we can't access full AP as it is not initialized yet.
        # Let's create a minimalistic AP and use it.
        ap = MiniAP(self.dp)
        ap.init()

        DEMCR_value = ap.read32(CortexM.DEMCR)

        # Halt on reset.
        ap.write32(CortexM.DEMCR, CortexM.DEMCR_VC_CORERESET)
        ap.write32(CortexM.DHCSR, CortexM.DBGKEY | CortexM.C_DEBUGEN)

        # Prevent disabling bus clock/power in low power modes.
        ap.write32(DBGMCU.CR, DBGMCU.CR_VALUE)

        self.dp.assert_reset(0)
        time.sleep(0.01)

        # Restore DEMCR original value.
        ap.write32(CortexM.DEMCR, DEMCR_value)

    def create_init_sequence(self):
        # STM32 under some low power/broken clock states doesn't allow AHP communication.
        # Low power modes are quite popular on stm32 (including MBed OS defaults).
        # 'attach' mode is broken by default, as STM32 can't be connected on low-power mode
        #  successfully without previous DBGMCU setup (It is not possible to write DBGMCU).
        # It is also not possible to run full pyOCD discovery code under-reset.
        #
        # As a solution we can setup DBGMCU under reset, halt core and release reset.
        # Unfortunately this code has to be executed _before_ discovery stage
        # and without discovery stage we don't have access to AP/Core.
        # As a solution we can create minimalistic AP implementation and use it
        # to setup core halt.
        # So the sequence for 'halt' connect mode will look like
        # -> Assert reset
        # -> Connect DebugPort
        # -> Setup MiniAp
        # -> Setup halt on reset
        # -> Enable support for debugging in low-power modes
        # -> Release reset
        # -> [Core is halted and reset is released]
        # -> Continue [discovery, create cores, etc]
        seq = super().create_init_sequence()
        if self.session.options.get('connect_mode') in ('halt', 'under-reset'):
            seq.insert_before('dp_init', ('assert_reset_for_connect', self.assert_reset_for_connect))
            seq.insert_after('dp_init', ('safe_reset_and_halt', self.safe_reset_and_halt))

        return seq

    def post_connect_hook(self):
        self.write32(DBGMCU.CR, DBGMCU.CR_VALUE)



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...core import exceptions
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (RomRegion, FlashRegion, RamRegion, MemoryMap)
from ...probe.swj import SWJSequenceSender
from ...probe.debug_probe import DebugProbe
from ...utility import mask

LOG = logging.getLogger(__name__)

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe00abe00,
    0xb085b5f0, 0x447e4e1e, 0x28017830, 0xf000d101, 0x2001f839, 0x70309004, 0x46384f15, 0xf00030f7,
    0x4604f8a3, 0x1c804813, 0xf89ef000, 0x46384605, 0xf89af000, 0x48109003, 0xf896f000, 0x480f9002,
    0xf892f000, 0x480b9001, 0xf88ef000, 0x47a04607, 0x607447a8, 0x980360b5, 0x980260f0, 0x98016130,
    0x61b76170, 0x70309804, 0xb0052000, 0x46c0bdf0, 0x00004552, 0x00005843, 0x00005052, 0x00004346,
    0x0000027a, 0x4c07b510, 0x7820447c, 0xd1062801, 0x47806960, 0x478069a0, 0x70202000, 0x2001bd10,
    0x46c0bd10, 0x000001f8, 0x44784805, 0x28007800, 0x2001d101, 0x48014770, 0x46c04770, 0x000070d0,
    0x000001d6, 0x4601b570, 0x447a4a0e, 0x28017810, 0x2301d10e, 0x2400071d, 0x46261b48, 0x42a94166,
    0x68d5d308, 0x041a0319, 0x47a823d8, 0xbd704620, 0xbd702001, 0x44784804, 0x4a042121, 0xf000447a,
    0x46c0f855, 0x000001b6, 0x00000126, 0x00000164, 0xb081b5f0, 0x4d10460b, 0x7829447d, 0xd10f2901,
    0x070e2101, 0x1b812400, 0x41674627, 0xd30b42b0, 0x4608692d, 0x461a4611, 0x462047a8, 0xbdf0b001,
    0x46202401, 0xbdf0b001, 0x44784804, 0x4a042121, 0xf000447a, 0x46c0f82b, 0x00000168, 0x000000d2,
    0x00000120, 0xd4d4de00, 0x2114b280, 0x1e898809, 0x2a00884a, 0x1d09d004, 0xd1f94282, 0x47708808,
    0x44784803, 0x4a03210e, 0xf000447a, 0x46c0f80f, 0x00000076, 0x000000c8, 0xd4d44770, 0x49024801,
    0x46c04770, 0x4d94efcf, 0x7847d224, 0xaf00b580, 0x2300b088, 0x4c079305, 0x93039404, 0x23019302,
    0xab069301, 0x91079300, 0x46689006, 0xf0004611, 0xdefef803, 0x00000244, 0xaf00b580, 0x9103b084,
    0x48049002, 0x48049001, 0x46689000, 0xffbaf7ff, 0x46c0defe, 0x00000244, 0x00000244, 0x636e7546,
    0x746f6e20, 0x756f6620, 0x7273646e, 0x616d2f63, 0x722e6e69, 0xd4d4d473, 0xd4d4d4d4, 0xd4d4d4d4,
    0x65747461, 0x2074706d, 0x73206f74, 0x72746275, 0x20746361, 0x68746977, 0x65766f20, 0x6f6c6672,
    0xd4d4d477, 0x00000199, 0x00000000, 0x00000001, 0x0000019d, 0x0000020a, 0x0000000b, 0x00000014,
    0x00000011, 0x0000020a, 0x0000000b, 0x00000053, 0x00000028, 0x0000020a, 0x0000000b, 0x00000058,
    0x0000002a, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000005,
    'pc_unInit': 0x20000089,
    'pc_program_page': 0x20000115,
    'pc_erase_sector': 0x200000c9,
    # 'pc_eraseAll': 0x200000ad, # not implemented yet

    'static_base' : 0x20000000 + 0x00000004 + 0x00000254,
    'begin_stack' : 0x20000500,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x0,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001100],   # Enable double buffering
    'min_program_length' : 256,

    # Relative region addresses and sizes
    'ro_start': 0x0,
    'ro_size': 0x254,
    'rw_start': 0x254,
    'rw_size': 0x30,
    'zi_start': 0x284,
    'zi_size': 0x1c,
}

def _parity32(value):
    parity = sum((value >> i) for i in range(32))
    return parity & 1

class RP2040Base(CoreSightTarget):
    """@brief Raspberry Pi RP2040.

    This device is very strange in that it as three DPs. The first two DPs each have a single AHB-AP
    for the two Cortex-M0+ cores. The third DP is a "Rescue DP" that has no APs, but the CDBGPWRUPREQ
    signal is repurposed as a rescue signal.
    """

    class Targetsel:
        """@brief DP TARGETEL values for each DP."""
        CORE_0 = 0x01002927
        CORE_1 = 0x11002927
        RESCUE_DP = 0xf1002927

    VENDOR = "Raspberry Pi"

    MEMORY_MAP = MemoryMap(
        RomRegion(  start=0,            length=0x4000,      name="bootrom",             ),
        FlashRegion(start=0x10000000,   length=0x1000000,   name="xip",
            sector_size=4096,
            page_size=256,
            algo=FLASH_ALGO,
            is_boot_memory=True),
        RomRegion(  start=0x11000000,   length=0x1000000,   name="xip_noalloc",           alias="xip"),
        RomRegion(  start=0x12000000,   length=0x1000000,   name="xip_nocache",           alias="xip"),
        RomRegion(  start=0x13000000,   length=0x1000000,   name="xip_noalloc_nocache",   alias="xip"),
        RamRegion(  start=0x20000000,   length=0x40000,     name="sram0_3",               ),
        RamRegion(  start=0x20040000,   length=0x2000,      name="sram4_5",               ),
        RamRegion(  start=0x21000000,   length=0x40000,     name="sram0_3_alias",         alias="sram0_3"),
        RamRegion(  start=0x51000000,   length=0x1000,      name="usbram",                ),
        )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)

        ## The TARGETSEL value to be used.
        self._core_targetsel = None

        # Disable the SWJ sequence. This performs a line reset, which causes all DPs to be selected. After
        # any line reset on a multi-drop SWD setup there must be a target selection sequence.
        session.options['dap_swj_enable'] = False

    def create_init_sequence(self):
        seq = super().create_init_sequence()

        seq.insert_before('load_svd', ('check_probe', self._check_probe)) \
            .insert_before('dp_init', ('select_core0', self._select_core))

        return seq

    def _check_probe(self):
        # Have to import here to avoid a circular import
        from ...probe.debug_probe import DebugProbe
        if DebugProbe.Capability.SWD_SEQUENCE not in self.session.probe.capabilities:
            raise exceptions.TargetSupportError("RP2040 requires a debug probe with SWD sequence capability")

    def _select_core(self):
        self.select_dp(self._core_targetsel)

    def select_dp(self, targetsel):
        """@brief Select the DP with the matching TARGETSEL."""
        probe = self.session.probe

        # Have to connect the probe first, or SWCLK will not be enabled.
        probe.connect(DebugProbe.Protocol.SWD)

        # First perform the dormant to SWD sequence. An SW-DPv2 implementing multi-drop SWD will cold
        # reset into dormant mode.
        swj = SWJSequenceSender(self.session.probe, True)
        swj.dormant_to_swd()

        # SWD line reset to activate all DPs.
        swj.line_reset()
        swj.idle_cycles(2)

        # Send multi-drop SWD target selection sequence to select the requested DP.
        probe.swd_sequence([
            # DP TARGETSEL write
            # output 8 cycles:
            #   - Start = 1
            #   - APnDP = 0
            #   - RnW = 0
            #   - A[2:3] = 2'b11
            #   - Parity = 0
            #   - Stop = 0
            #   - Park = 1
            # -> LSB first, that's 0b10011001 or 0x99
            (8, 0x99),

            # 5 cycles with SWDIO as input
            (5,),

            # DP TARGETSEL value
            # output 32 + 1 cycles
            (33, targetsel | mask.parity32_high(targetsel)),

            # 2 idle cycles
            (2, 0x00),
            ])

        DP_IDR = 0x00
        dpidr = probe.read_dp(DP_IDR)
        LOG.debug("DP IDR after writing TARGETSEL: 0x%08x", dpidr)

        probe.write_dp(0x8, 0x2) # DPBANKSEL=2 to select TARGETID
        targetid = probe.read_dp(0x4)
        LOG.debug("DP TARGETID: 0x%08x", targetid)

        probe.write_dp(0x8, 0x3) # DPBANKSEL=3 to select DLPIDR
        dlpidr = probe.read_dp(0x4)
        LOG.debug("DP DLPIDR: 0x%08x", dlpidr)

        probe.write_dp(0x8, 0x0) # restore DPBANKSEL=0

class RP2040Core0(RP2040Base):
    """@brief RP2040 target for core 0."""

    def __init__(self, session):
        super().__init__(session)
        self._core_targetsel = self.Targetsel.CORE_0

class RP2040Core1(RP2040Base):
    """@brief RP2040 target for core 1."""

    def __init__(self, session):
        super().__init__(session)
        self._core_targetsel = self.Targetsel.CORE_1




# pyOCD debugger
# Copyright (c) 2017 NXP
# Copyright (c) 2006-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...core import exceptions
from ...core.memory_map import MemoryType
from ...coresight.coresight_target import CoreSightTarget
from ...coresight.cortex_m import CortexM

LOG = logging.getLogger(__name__)

class IMXRT(CoreSightTarget):
    VENDOR = "NXP"

    def create_init_sequence(self):
        seq = super().create_init_sequence()
        seq.wrap_task('discovery',
            lambda seq: seq.replace_task('create_cores', self.create_cores)
            )
        return seq

    def create_cores(self):
        try:
            core = CortexM7_IMXRT(self.session, self.aps[0], self.memory_map, 0)
            core.default_reset_type = self.ResetType.SW_VECTRESET
            self.aps[0].core = core
            core.init()
            self.add_core(core)
        except KeyError:
            LOG.error("No core-0 were discovered")



class CortexM7_IMXRT(CortexM):

    # System Control Space(SRC)
    SRC_SBMR1 = 0x400F8004
    SRC_SBMR2 = 0x400F801C

    FPB_CTRL = 0xE0002000
    FPB_COMP0 = 0xE0002008

    BOOT_MODES = {
        0x00: "Boot From Fuses",
        0x01: "Serial Download Mode",
        0x02: "Internal Boot",
        0x03: "Reserved"
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.get_boot_mode()

    def get_boot_mode(self):
        # Read Boot Mode
        # SBMR2: Bit 25..24:
        # BOOT_MODE[1:0]: 00b - Boot From Fuses
        #                 01b - Serial Downloader
        #                 10b - Internal Boot
        #                 11b - Reserved
        bootmode = (self.read_memory(CortexM7_IMXRT.SRC_SBMR2) & 0x03000000) >> 24
        LOG.info("IMXRT Boot Mode: %s" % self.BOOT_MODES[bootmode])
        return bootmode

    def get_boot_device(self):
        # Read Boot Device
        # Boot Device: 0000b - Serial NOR boot via FlexSPI
        #              001xb - SD boot via uSDHC
        #              10xxb - eMMC/MMC boot via uSDHC
        #              01xxb - SLC NAND boot via SEMC
        #              0001b - Parallel NOR boot via SEMC
        #              11xxb - Serial NAND boot via FlexSPI
        bootdevice = (self.read_memory(CortexM7_IMXRT.SRC_SBMR1) & 0x000000F0) >> 4
        LOG.info("IMXRT Boot Device: %x" % bootdevice)
        return bootdevice

    def _get_flash_vector_addr(self):
        mem = self.memory_map.get_boot_memory()
        if mem and mem.type == MemoryType.FLASH:
            return mem.start + 0x1004
        return None

    def set_reset_catch(self, reset_type=None):
        self.did_normal_reset_catch = True
        bootmode = self.get_boot_mode()
        bootdevice = self.get_boot_device()

        # boot from flexspi_nor
        if bootmode == 2 and bootdevice == 0 and \
            reset_type not in (self.ResetType.SW_SYSRESETREQ, self.ResetType.SW_VECTRESET):
            # Disable Reset Vector Catch in DEMCR
            value = self.read_memory(CortexM.DEMCR)
            self.write_memory(CortexM.DEMCR, (value & (~0x00000001)))
            vectable_addr = self._get_flash_vector_addr()
            LOG.debug("vectable_addr: %s", hex(vectable_addr) if (vectable_addr is not None) else "None")
            vectable = None
            imageentry = None

            if vectable_addr:
                try:
                    # Read user Image Vector Table address
                    vectable = self.read_memory(vectable_addr)
                    if vectable and vectable != 0xFFFFFFFF:
                        # Read user image entry point and clear Thumb bit
                        imageentry = self.read_memory(vectable + 4) & (~0x00000001)
                except (AssertionError, exceptions.TransferFaultError):
                    pass

                if imageentry and imageentry != 0xFFFFFFFF:
                    LOG.debug("vectable: %s, imageentry: %s" % (vectable, imageentry))
                    # Program FPB Comparator 0 to user image entry point
                    self.write_memory(CortexM7_IMXRT.FPB_COMP0, (imageentry | 1))
                    # Enable FPB (FPB_CTRL = FPB_KEY|FPB_ENABLE)
                    self.write_memory(CortexM7_IMXRT.FPB_CTRL, 0x00000003)
                    LOG.debug("enable fpb")
                    self.did_normal_reset_catch = False
                    return

        # normal reset catch
        LOG.debug("normal_set_reset_catch")
        self.did_normal_reset_catch = True
        super().set_reset_catch(reset_type)

    def clear_reset_catch(self, reset_type=None):
        if self.did_normal_reset_catch:
            super().clear_reset_catch(reset_type)
        else:
            # Disable Reset Vector Catch in DEMCR
            value = self.read_memory(CortexM.DEMCR)
            self.write_memory(CortexM.DEMCR, (value& (~0x00000001)))
            # Clear BP0 and FPB
            self.write_memory(CortexM7_IMXRT.FPB_COMP0, 0);                        # Clear BP0
            self.write_memory(CortexM7_IMXRT.FPB_CTRL, 0x00000002);                # Disable FPB
            LOG.debug("clear fpb")



# pyOCD debugger
# Copyright (c) 2015-2017 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from ...core.target import Target

class Breakpoint:
    def __init__(self, provider):
        self.type: Target.BreakpointType = Target.BreakpointType.HW
        self.enabled: bool = False
        self.addr: int = 0
        self.original_instr: int = 0
        self.provider: BreakpointProvider = provider

    def __repr__(self) -> str:
        return "<%s@0x%08x type=%s addr=0x%08x>" % (self.__class__.__name__, id(self), self.type.name, self.addr)

class BreakpointProvider:
    """@brief Abstract base class for breakpoint providers."""
    def init(self) -> None:
        raise NotImplementedError()

    @property
    def bp_type(self) -> Target.BreakpointType:
        raise NotImplementedError()

    @property
    def do_filter_memory(self) -> bool:
        return False

    @property
    def available_breakpoints(self) -> int:
        raise NotImplementedError()

    def can_support_address(self, addr: int) -> bool:
        raise NotImplementedError()

    def find_breakpoint(self, addr: int) -> Optional[Breakpoint]:
        raise NotImplementedError()

    def set_breakpoint(self, addr: int) -> Optional[Breakpoint]:
        raise NotImplementedError()

    def remove_breakpoint(self, bp: Breakpoint) -> None:
        raise NotImplementedError()

    def filter_memory(self, addr: int, size: int, data: int) -> int:
        return data

    def flush(self) -> None:
        pass



